from pytorch2keras.converter import pytorch_to_keras
from torch import nn
from torch.nn import functional as F
import torch

import torch
import torch.nn as nn
import torch.nn.functional as F


class Bottleneck(nn.Module):
    def __init__(self, inp_c, out_c, kernel_size, stride, t=1):
        assert stride in [1, 2], 'stride must be either 1 or 2'
        super().__init__()
        self.residual = stride == 1 and inp_c == out_c
        # pad = kernel_size // 2
        # self.reflection_pad = nn.ReflectionPad2d(pad)
        self.conv1 = nn.Conv2d(inp_c, t*inp_c, 1, 1, bias=False)
        self.in1 = nn.InstanceNorm2d(t*inp_c, affine=True)
        self.conv2 = nn.Conv2d(t*inp_c, t*inp_c, kernel_size, stride,
                               groups=t*inp_c, bias=False, padding=kernel_size//2)
        self.in2 = nn.InstanceNorm2d(t*inp_c, affine=True)
        self.conv3 = nn.Conv2d(t*inp_c, out_c, 1, 1, bias=False)
        self.in3 = nn.InstanceNorm2d(out_c, affine=True)

    def forward(self, x):
        out = F.relu6(self.in1(self.conv1(x)))
        # out = self.reflection_pad(out)
        out = F.relu6(self.in2(self.conv2(out)))
        out = self.in3(self.conv3(out))
        if self.residual:
            out = x + out
        return out


class UpsampleConv(nn.Module):
    def __init__(self, inp_c, out_c, kernel_size, stride, upsample=2):
        super().__init__()
        if upsample:
            self.upsample=upsample
            # self.upsample = nn.Upsample(mode='bilinear', scale_factor=upsample, align_corners=False)
            self.upsample = nn.Upsample(mode='nearest', scale_factor=2)
        else:
            self.upsample = None
        self.conv1 = Bottleneck(inp_c, out_c, kernel_size, stride)

    def forward(self, x):
        x_in = x
        if self.upsample is not None:
            x_in = self.upsample(x_in)
        out = F.relu(self.conv1(x_in))
        return out


class TransformerMobileNet(nn.Module):
    def __init__(self):
        super().__init__()
        # Conv Layers
        # self.reflection_pad = nn.ReflectionPad2d(9//2)
        self.conv1 = nn.Conv2d(3, 32, kernel_size=9, stride=1, bias=False, padding=9//2)
        self.in1 = nn.InstanceNorm2d(32, affine=True)
        self.conv2 = Bottleneck(32, 64, kernel_size=3, stride=2)
        self.conv3 = Bottleneck(64, 128, kernel_size=3, stride=2)
        # Residual Layers
        self.res1 = Bottleneck(128, 128,  3, 1)
        self.res2 = Bottleneck(128, 128,  3, 1)
        self.res3 = Bottleneck(128, 128,  3, 1)
        self.res4 = Bottleneck(128, 128,  3, 1)
        self.res5 = Bottleneck(128, 128,  3, 1)
        # Upsampling Layers
        self.upconv1 = UpsampleConv(128, 64, kernel_size=3, stride=1)
        self.upconv2 = UpsampleConv(64, 32, kernel_size=3, stride=1)
        self.conv4 = nn.Conv2d(32, 3, kernel_size=9, stride=1, bias=False, padding=9//2)

    def forward(self, x):
        # out = self.reflection_pad(x)
        out = F.relu(self.in1(self.conv1(x)))
        out = self.conv2(out)
        # out = self.conv3(out)
        # out = self.res1(out)
        # out = self.res2(out)
        # out = self.res3(out)
        # out = self.res4(out)
        # out = self.res5(out)
        # out = self.upconv1(out)
        # out = self.upconv2(out)
        # out = self.conv4(out)
        return out

# if __name__ == "__main__":
#     from torchsummary import summary
#     model = TransformerMobileNet()
#     model.to('cuda')
#     summary(model, (3, 256, 256))
#     model.load_state_dict(torch.load('saved_model/release.pth'))
#     torch.onnx.export(model, torch.randn(1, 3, 256, 256).cuda(), 'models/style1.onnx')
import numpy as np
with torch.no_grad():
    test_net = TransformerMobileNet()
    test_net.eval()
    dummy_input = torch.ones([1, 3, 256, 256])
    dummy_output = test_net(dummy_input)

    print(test_net(dummy_input))
    t_model = pytorch_to_keras(test_net, dummy_input, [(3, 256, 256)], change_ordering=True, verbose=True)
    t_model.summary()
    t_model.save('t_model.h5')

    k_pred = t_model.predict(np.ones((1, 256, 256, 3)))
    print(k_pred - dummy_output.permute(0, 2, 3, 1).numpy())


import torch
import torch.nn as nn
import numpy as np
from torch.autograd import Variable

class PytorchModel(nn.Module):
    def __init__(self):
        super(PytorchModel, self).__init__()
        self.bn = nn.BatchNorm2d(1, momentum=0.5)

    def forward(self, x):
        x = self.bn(x)
        return x

pytorch_model = PytorchModel()
pytorch_model.eval()

x_data = np.random.uniform(0, 1, size=(1, 1, 224, 224))
x = Variable(torch.Tensor(x_data))

torch_out = torch.onnx._export(pytorch_model, x, "model.onnx", export_params=True)


import numpy as np
import torch
import torch.nn as nn

# import tensorflow as tf
# import keras
# config = tf.ConfigProto(intra_op_parallelism_threads=1, 
#                         inter_op_parallelism_threads=1, 
#                         allow_soft_placement=False, 
#                         device_count = {'CPU': 1})
# session = tf.Session(config=config)
# keras.backend.set_session(session)

from torch.autograd import Variable
from pytorch2keras.converter import pytorch_to_keras


class pixel_shuffle(nn.Module):
    def __init__(self, scale_factor):
        super(pixel_shuffle, self).__init__()
        self.scale_factor = scale_factor

    def forward(self, input):
        scale_factor = self.scale_factor
        _, in_channels, in_height, in_width = input.shape

        in_channels = int(in_channels)
        in_height = int(in_height)
        in_width = int(in_width)

        out_channels = in_channels // (scale_factor * scale_factor)
        out_height = in_height * scale_factor
        out_width = in_width * scale_factor

        if scale_factor >= 1:
            input_view = input.view([-1, out_channels, scale_factor, scale_factor, in_height, in_width])
            shuffle_out = input_view#.permute(0, 1, 4, 2, 5, 3)

        return shuffle_out.contiguous() #.view([-1, out_channels, out_height, out_width])

# graph(%0 : Float(1, 4, 64, 64)) {
#   %2 : Tensor = onnx::Constant[value= -1   4   1   1  64  64 [ CPULongType{6} ]](), scope: LayerTest/pixel_shuffle[psh]
#   %3 : Float(1, 4, 1, 1, 64, 64) = onnx::Reshape(%0, %2), scope: LayerTest/pixel_shuffle[psh]
#   %5 : Float(1, 4, 64!, 1!, 64!, 1!) = onnx::Transpose[perm=[0, 1, 4, 2, 5, 3]](%3), scope: LayerTest/pixel_shuffle[psh]
#   %7 : Tensor = onnx::Constant[value= -1   4  64  64 [ CPULongType{4} ]](), scope: LayerTest/pixel_shuffle[psh]
#   %8 : Float(1, 4, 64, 64) = onnx::Reshape(%5, %7), scope: LayerTest/pixel_shuffle[psh]
#   return (%8);
# }

# graph(%0 : Float(1, 4, 64, 64)) {
#   %2 : Tensor = onnx::Constant[value= -1   1   1   4  64  64 [ CPULongType{6} ]](), scope: LayerTest/PixelShuffle[psh]
#   %3 : Tensor = onnx::Reshape(%0, %2), scope: LayerTest/PixelShuffle[psh]
#   %4 : Tensor = onnx::Transpose[perm=[0, 1, 4, 2, 5, 3]](%3), scope: LayerTest/PixelShuffle[psh]
#   %5 : Tensor = onnx::Constant[value= -1   4  64  64 [ CPULongType{4} ]](), scope: LayerTest/PixelShuffle[psh]
#   %6 : Float(1, 4, 64, 64) = onnx::Reshape(%4, %5), scope: LayerTest/PixelShuffle[psh]
#   return (%6);
# }



class LayerTest(nn.Module):
    def __init__(self):
        super(LayerTest, self).__init__()
        self.psh = pixel_shuffle(1)

    def forward(self, x):
        x = self.psh(x.contiguous())
        return x


def check_error(output, k_model, input_np, epsilon=1e-5):
    pytorch_output = output.data.numpy()
    keras_output = k_model.predict(input_np)

    error = np.max(pytorch_output - keras_output)
    print('Error:', error)

    assert error < epsilon
    return error


def freeze_session(session, keep_var_names=None, output_names=None, clear_devices=True):
    """
    Freezes the state of a session into a pruned computation graph.

    Creates a new computation graph where variable nodes are replaced by
    constants taking their current value in the session. The new graph will be
    pruned so subgraphs that are not necessary to compute the requested
    outputs are removed.
    @param session The TensorFlow session to be frozen.
    @param keep_var_names A list of variable names that should not be frozen,
                          or None to freeze all the variables in the graph.
    @param output_names Names of the relevant graph outputs.
    @param clear_devices Remove the device directives from the graph for better portability.
    @return The frozen graph definition.
    """
    from tensorflow.python.framework.graph_util import convert_variables_to_constants
    graph = session.graph
    with graph.as_default():
        freeze_var_names = \
            list(set(v.op.name for v in tf.global_variables()).difference(keep_var_names or []))
        output_names = output_names or []
        output_names += [v.op.name for v in tf.global_variables()]
        input_graph_def = graph.as_graph_def()
        if clear_devices:
            for node in input_graph_def.node:
                node.device = ""
        frozen_graph = convert_variables_to_constants(session, input_graph_def,
                                                      output_names, freeze_var_names)
        return frozen_graph




if __name__ == '__main__':
    max_error = 0
    for i in range(10):
        psh_1 = pixel_shuffle(2)
        # psh_2 = nn.PixelShuffle(2)

        model = LayerTest()
        model.eval()

        input_np = np.random.uniform(0, 1, (1, 4, 64, 64))
        input_var = Variable(torch.FloatTensor(input_np))
        
        # print(psh_1(input_var) - psh_2(input_var))

        output = model(input_var)

        k_model = pytorch_to_keras(model, input_var, (4, 64, 64,), verbose=True)
        k_model.summary()

        from keras import backend as K
        import tensorflow as tf
        frozen_graph = freeze_session(K.get_session(),
                                      output_names=[out.op.name for out in k_model.outputs])

        tf.train.write_graph(frozen_graph, ".", "my_model.pb", as_text=False)
        print([i for i in k_model.outputs])


        print(k_model.to_json())
        k_model.save_weights('test.h5')
        error = check_error(output, k_model, input_np)
        if max_error < error:
            max_error = error

    print('Max error: {0}'.format(max_error))
